create or replace PROCEDURE SP_ALTA_LECTURA (
   p_comptador   IN  LECTURA.comptador%TYPE,
   p_data        IN  LECTURA.data%TYPE,
   p_lectura_KWh IN  LECTURA.lectura_KWh%TYPE,
   p_tipus       IN  LECTURA.tipus%TYPE, 
   RSP           OUT VARCHAR2
)
IS
/******************************************************************************
   Nom:                 SP_ALTA_LECTURA
   Propsit:            afegeix una lectura a la taula LECTURA si aquesta es la ms recent.
                        (excepte per la primera lectura, sempre ha d'haver 
                         almenys una lectura registrada al mes anterior)
                   
   Parmetres entrada:  p_comptador, p_data, p_lectura_KWh, p_tipus             
   Parmetres sortida:  RSP
   Retorna:             'OK' si l'execuci ha finalitzat amb xit
                        'ERROR: tipus d'error' si ha fracassat
******************************************************************************/
param_entrada  VARCHAR2(500);
existeix     INTEGER;
comptadors   NUMBER;
err_msg      VARCHAR2 (500);
camp_nul     VARCHAR2(20);
consum       NUMBER;
lec_anterior NUMBER;
codi_cprod   NUMBER;
codi_linia   NUMBER;
any_lectura  NUMBER;
ultima_data  DATE;
E_NUL        EXCEPTION;
E_FK         EXCEPTION;
E_LEC_VALOR  EXCEPTION;
E_LEC_DATA   EXCEPTION;
E_MES_NO_LEC EXCEPTION;
E_BAIXA      EXCEPTION;

type r_cursor IS REF CURSOR;
CUR r_cursor;

BEGIN
   
   param_entrada:= 'p_comptador: '||p_comptador||
                 ', p_data: '||p_data||
                 ', p_lectura_KWh: '||p_lectura_KWh||
                 ', p_tipus: '||p_tipus;
                 
  --verifica que els camps obligatoris no siguin nuls
   IF (p_comptador IS NULL) THEN 
     camp_nul:= 'comptador';
     RAISE E_NUL;
   ELSIF (p_data IS NULL) THEN 
     camp_nul:= 'data';
     RAISE E_NUL;
   ELSIF (p_lectura_KWh IS NULL) THEN 
     camp_nul:= 'lectura_KWh';
     RAISE E_NUL;  
   ELSIF (p_tipus IS NULL) THEN 
     camp_nul:= 'tipus';
     RAISE E_NUL;  
   END IF; 

  --verifica que el comptador existeix a la base de dades.
   SELECT COUNT (*)
   INTO existeix
   FROM COMPTADOR
   WHERE codi = p_comptador;

   IF (existeix = 0) THEN
     RAISE E_FK;
   END IF;      
   
   
  --verifica que l'estat del comptador sigui ALTA.
   SELECT COUNT (*)
   INTO existeix
   FROM COMPTADOR
   WHERE codi = p_comptador AND estat = 'ALTA';

   IF (existeix = 0) THEN
     RAISE E_BAIXA;
   END IF;      
      
   
  --calcula l'ultima data
   SELECT MAX(data) 
   INTO ultima_data 
   FROM LECTURA 
   WHERE comptador = p_comptador;
      
  --verifica que si no es tracta de la primera lectura, s'ha d'haver enregistrat
  --almenys una lectura al mes anterior.
   IF ( (EXTRACT(year FROM ultima_data) = EXTRACT(year FROM p_data) 
         AND EXTRACT(month FROM ultima_data) < EXTRACT(month FROM p_data)-1) 
         OR
        (EXTRACT(year FROM ultima_data) < EXTRACT(year FROM p_data)
         AND (EXTRACT(month FROM ultima_data)<>12 OR EXTRACT(month FROM p_data) <> 1 
              OR EXTRACT(year FROM p_data) > EXTRACT(year FROM ultima_data)+1))
      ) THEN
        RAISE E_MES_NO_LEC;
   END IF; 
    
  --calcula lectura anterior
   SELECT MAX(lectura_KWh) 
   INTO lec_anterior 
   FROM LECTURA 
   WHERE comptador = p_comptador;
   
   --verifica que la lectura sigui la ms recent
   IF (p_data <= ultima_data)  THEN
     RAISE E_LEC_DATA;
   END IF;  
   
  --verifica que la lectura sigui major o igual que l'anterior
   IF (p_lectura_kwh < lec_anterior)  THEN
     RAISE E_LEC_VALOR;
   END IF;
         
  --calcula el cosum  fins a la data de la lectura respecte a l'ultima lectura.
   SELECT p_lectura_KWh - NVL( MAX(lectura_KWh), 0)
   INTO consum
   FROM LECTURA
   WHERE comptador=p_comptador;
   
  -- inserta la nova lectura
   INSERT INTO LECTURA (comptador, data, lectura_KWh, tipus, consum_KWh)
   VALUES (p_comptador, p_data, p_lectura_KWh, UPPER(p_tipus), consum);
   
  -- inserci amb xit
   RSP:= 'OK';
   SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST1 ---
  ------------------------------------------------------------------------------

  --Calcula el codi de la central de producci de la que depen el comptador. 
   SELECT DISTINCT con.central_produccio
   INTO codi_cprod
   FROM DISTRIBUCIO dist JOIN CONNEXIO con ON con.central_distribucio = dist.central_distribucio
   WHERE dist.comptador = p_comptador 
   AND dist.servei_actiu='SI' AND con.servei_actiu='SI';

  -- calcula si existeix una entrada a la taula per la central de produccio de la que depen
  -- el comptador de la lectura.
   SELECT COUNT (*)
   INTO existeix
   FROM EST1_CPROD_CONSUM
   WHERE central_produccio = codi_cprod;
   
  --Si no existeix l'entrada, l'afegeix a EST1 
   IF (existeix = 0) THEN 
     SP_ALTA_EST1(codi_cprod, consum);
     
  --Si existeix actualitza la entrada.
   ELSE
   SP_MODIFICACIO_EST1(codi_cprod);
   END IF;
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST2 ---
--(Si un comptador es connecta a dues o ms lnies de comunicaci, computa el consum a totes,
-- encara que no siguin connexions actives)--
  ------------------------------------------------------------------------------

  --calcula l'any de la lectura
   any_lectura:= EXTRACT(year FROM p_data);

  --Calcula el codi de la lnia de comunicaci de la que depen el comptador.
   OPEN CUR FOR
   SELECT DISTINCT con.linia_comunicacio
   FROM DISTRIBUCIO dist JOIN CONNEXIO con ON con.central_distribucio = dist.central_distribucio
   WHERE dist.comptador = p_comptador AND con.servei_actiu='SI';

  --Per cada lnia relacionada:
   LOOP
     FETCH CUR INTO codi_linia;
     EXIT WHEN CUR%NOTFOUND;
      --calcula si existeix una entrada a la taula per la lnia de comunicac a l'any de la lectura.
       SELECT COUNT (*)
       INTO existeix
       FROM EST2_MITJA_CONSUM_LINIA
       WHERE linia = codi_linia AND anyo = any_lectura ;

      --Si no existeix l'entrada, l'afegeix a EST2
       IF (existeix = 0) THEN 
         SP_ALTA_EST2(codi_linia, any_lectura, consum);
         
      --Si existeix actualitza la entrada.
       ELSE
         SP_MODIFICACIO_EST2(codi_linia, any_lectura);
         
       END IF;
   END LOOP;  
   CLOSE CUR;  

  
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST3 ---
  ------------------------------------------------------------------------------

  -- calcula si existeix una entrada a la taula 
   SELECT COUNT (*)
   INTO existeix
   FROM EST3_LINIA_MES_CARREGA;

   
  --Si no existeix l'entrada, l'afegeix a EST3 
   IF (existeix = 0) THEN 
     SP_ALTA_EST3(codi_linia, consum);
     
  --Si existeix actualitza la entrada.
   ELSE
     SP_MODIFICACIO_EST3();
   END IF;
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST4 ---
  ------------------------------------------------------------------------------

  -- calcula si existeix una entrada a la taula per l'any de la lectura
   SELECT COUNT (*)
   INTO existeix
   FROM EST4_LINIES_MES_50
   WHERE anyo = any_lectura;
   
  --Si no existeix l'entrada, l'afegeix a EST4 
   IF (existeix = 0) THEN 
     SP_ALTA_EST4(any_lectura);
     
  --Si existeix actualitza la entrada.
   ELSE
   SP_MODIFICACIO_EST4(any_lectura); 
   END IF;
      
      
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST6 ---
  ------------------------------------------------------------------------------

  -- calcula si ja existeixen 10 comptadors a la taula EST6
   SELECT COUNT (*)
   INTO comptadors
   FROM EST6_COMPT_MES_CONSUM;
   
  -- calcula si el comptador existeix a la taula EST6
   SELECT COUNT (*)
   INTO existeix
   FROM EST6_COMPT_MES_CONSUM;   
   
  --Si el comptador no existeix a EST6 i la taula te menys de 10 entrades, afegeix el comptador
   IF (comptadors < 10 AND existeix = 0) THEN 
     SP_ALTA_EST6(p_comptador, consum);
     
   --En cas contrari actualitza la taula
   ELSE
   SP_MODIFICACIO_EST6(); 
   END IF;
   
   ------------------------------------------------------------------------------
                    -- ACTUALITZA EST7 ---
  ------------------------------------------------------------------------------

  -- calcula si existeix una entrada a la taula EST7
   SELECT COUNT (*)
   INTO existeix
   FROM EST7_MITJA_CONSUM_CLIENTS;

   
  --Si no existeix l'entrada, l'afegeix a EST7 
   IF (existeix = 0) THEN 
     SP_ALTA_EST7(consum);
     
  --Si existeix actualitza la entrada.
   ELSE
     SP_MODIFICACIO_EST7();
   END IF;    
   
   
   
  -------------------------F MODUL ESTADSTIC ---------------------------------   
   COMMIT;
EXCEPTION
   WHEN E_NUL THEN
        RSP:= 'ERROR: el camp ' || camp_nul || ' no pot ser nul';
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
               
   WHEN E_FK THEN
        RSP:= 'ERROR: el comptador no existeix';    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_BAIXA THEN
        RSP:= 'ERROR: el comptador est donat de baixa';    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);        

   WHEN E_LEC_DATA THEN
        RSP:= 'ERROR: el valor del camp data ha de ser posterior que ' || ultima_data; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_LEC_VALOR THEN
        RSP:= 'ERROR: el valor del camp lectura_KWh ha de ser major o igual que ' || lec_anterior;   
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_MES_NO_LEC THEN
        RSP:= 'ERROR: no es pot enregistrar la lectura. Ha d''existir almenys una lectura al mes anterior';
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
     
   WHEN OTHERS THEN
        ROLLBACK;
        IF (SQLCODE = -00001) THEN 
          RSP:= 'ERROR: ja existeix una lectura amb mateix: comptador, data'; 
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
          
        ELSIF (SQLCODE = -02290) THEN 
          RSP:= 'ERROR: el valor del camp tipus ha de ser TELEMATIC/PRESENCIAL';
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
          
        ELSE
          err_msg := SUBSTR (SQLERRM, 1, 500);
          RSP:=    'ERROR: ' || err_msg;
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_LECTURA', param_entrada, 'RSP: '||RSP);
        END IF;
              
END SP_ALTA_LECTURA;