create or replace PROCEDURE SP_MODIFICACIO_LECTURA (
   p_codi        IN  LECTURA.codi%TYPE,
   p_data        IN  LECTURA.data%TYPE,
   p_lectura_KWh IN  LECTURA.lectura_KWh%TYPE,
   p_tipus       IN  LECTURA.tipus%TYPE, 
   RSP           OUT VARCHAR2
)
IS
/******************************************************************************
   Nom:                 SP_MODIFICACIO_LECTURA
   Propsit:            modifica les dades d'una lectura a la taula LECTURA
                        si es tracta de la lectura ms recent d'un comptador.
                        
   Parmetres entrada:  p_codi, p_data, p_lectura_KWh, p_tipus             
   Parmetres sortida:  RSP
   Retorna:             'OK' si l'execuci ha finalitzat amb xit
                        'ERROR: tipus d'error' si ha fracassat
******************************************************************************/
param_entrada  VARCHAR2(500);
existeix       INTEGER;
err_msg        VARCHAR2 (200);
camp_nul       VARCHAR2(20);
codi_comptador NUMBER;
data_lectura   DATE;
ultima_data    DATE;
data_anterior  DATE;
lec_anterior   NUMBER;
consum         NUMBER;
codi_cprod     NUMBER;
codi_linia     NUMBER;
any_lectura    NUMBER;
i              INTEGER:=1;
E_NUL          EXCEPTION;
E_NO_EXISTEIX  EXCEPTION;
E_LEC_VALOR    EXCEPTION;
E_LEC_DATA     EXCEPTION;
E_MES_NO_LEC   EXCEPTION;
E_NO_MOD       EXCEPTION;

type r_cursor IS REF CURSOR;
CUR r_cursor;

BEGIN
   
   param_entrada:= 'p_codi: '||p_codi||
                 ', p_data: '||p_data||
                 ', p_lectura_KWh: '||p_lectura_KWh||
                 ', p_tipus: '||p_tipus;
   
   
  --verifica que els camps obligatoris no siguin nuls
   IF (p_codi IS NULL) THEN 
     camp_nul:= 'codi';
     RAISE E_NUL;
   ELSIF (p_data IS NULL) THEN 
     camp_nul:= 'data';
     RAISE E_NUL;    
   ELSIF (p_lectura_KWh IS NULL) THEN 
     camp_nul:= 'lectura_KWh';
     RAISE E_NUL;  
   ELSIF (p_tipus IS NULL) THEN 
     camp_nul:= 'tipus';
     RAISE E_NUL;  
   END IF; 

  --verifica que la lectura existeix
   SELECT COUNT (*)
   INTO existeix
   FROM LECTURA
   WHERE codi = p_codi ;

   IF (existeix = 0) THEN
     RAISE E_NO_EXISTEIX;
   END IF;      

  --calcula el comptador i data de la lectura existent
   SELECT comptador, data
   INTO codi_comptador, data_lectura
   FROM LECTURA 
   WHERE codi = p_codi;  

  --calcula l'ltima data
   SELECT MAX(data)
   INTO ultima_data
   FROM LECTURA
   WHERE comptador = codi_comptador;

  --verifica que es tracta de la lectura ms recent
   IF ( data_lectura <> ultima_data ) THEN
     RAISE E_NO_MOD;
   END IF;    
  
  --calcula lectura anterior
   SELECT MAX(lectura_KWh) 
   INTO lec_anterior 
   FROM LECTURA 
   WHERE comptador = codi_comptador AND data < data_lectura;
   
  --verifica que la lectura sigui major o igual que l'anterior
   IF (p_lectura_kwh < lec_anterior)  THEN
     RAISE E_LEC_VALOR;
   END IF;   
   
  --calcula la data de la lectura anterior
   SELECT MAX(data)
   INTO data_anterior
   FROM LECTURA
   WHERE comptador = codi_comptador AND data < data_lectura;   
    
   --verifica que la nova data de lectura sigui la ms recent
   IF (p_data <= data_anterior)  THEN
     RAISE E_LEC_DATA;
   END IF;    
   
  --verifica que si no es tracta de la primera lectura, s'ha d'haver enregistrat
  --almenys una lectura al mes anterior.
   IF ( (EXTRACT(year FROM data_anterior) = EXTRACT(year FROM p_data) 
         AND EXTRACT(month FROM data_anterior) < EXTRACT(month FROM p_data)-1) 
         OR
        (EXTRACT(year FROM data_anterior) < EXTRACT(year FROM p_data)
         AND (EXTRACT(month FROM data_anterior)<>12 OR EXTRACT(month FROM p_data) <> 1 
              OR EXTRACT(year FROM p_data) > EXTRACT(year FROM data_anterior)+1))
      ) THEN
        RAISE E_MES_NO_LEC;
   END IF; 
  

          
  --calcula el cosum  fins a la data de la lectura respecte a l'ultima lectura.
   SELECT p_lectura_KWh - NVL(lectura_KWh, 0)
   INTO consum
   FROM LECTURA
   WHERE comptador=codi_comptador AND data = data_anterior ;   
    
  -- modifica les dades de la lectura
   UPDATE LECTURA 
   SET data = p_data,
       lectura_KWh = p_lectura_kwh, 
       tipus = UPPER(p_tipus),
       consum_KWh = consum
   WHERE codi = p_codi;   
   
  -- finalitza amb xit
   RSP:= 'OK';
   SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST1 ---
  ------------------------------------------------------------------------------
   
  --Calcula el codi de la central de producci de la que depen el comptador. 
   SELECT DISTINCT con.central_produccio
   INTO codi_cprod
   FROM DISTRIBUCIO dist JOIN CONNEXIO con ON con.central_distribucio = dist.central_distribucio
   WHERE dist.comptador = codi_comptador 
   AND dist.servei_actiu='SI' AND con.servei_actiu='SI';
    
  --Actualitza l'entrada de la central de producci a EST1.
   SP_MODIFICACIO_EST1(codi_cprod);
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST2 ---
  ------------------------------------------------------------------------------
   
  --calcula l'any de la lectura
   any_lectura:= EXTRACT(year FROM p_data);

  --Calcula el codi de la lnia de comunicaci de la que depen el comptador.
   OPEN CUR FOR
   SELECT DISTINCT con.linia_comunicacio
   FROM DISTRIBUCIO dist JOIN CONNEXIO con ON con.central_distribucio = dist.central_distribucio
   WHERE dist.comptador = codi_comptador AND con.servei_actiu='SI';
      
  --Per cada lnia relacionada:
   LOOP
     FETCH CUR INTO codi_linia;
     EXIT WHEN CUR%NOTFOUND;

      --actualitza la entrada a la taula EST2.
       SP_MODIFICACIO_EST2(codi_linia, any_lectura);

       i:= i+1;
   END LOOP;  
   CLOSE CUR;     
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST3 ---
  ------------------------------------------------------------------------------
     
  --Actualitza l'entrada de la central de producci a EST3.
   SP_MODIFICACIO_EST3();


  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST4 ---
  ------------------------------------------------------------------------------
     
  --Actualitza l'entrada de l'any a EST4.
   SP_MODIFICACIO_EST4(any_lectura);
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST6 ---
  ------------------------------------------------------------------------------
     
  --Actualitza l'entrada de la central de producci a EST6.
   SP_MODIFICACIO_EST6();   
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST7 ---
  ------------------------------------------------------------------------------
     
  --Actualitza l'entrada de la central de producci a EST7.
   SP_MODIFICACIO_EST7();     
------------------------------------------------------------------------------   
   
   COMMIT;
EXCEPTION
   WHEN E_NUL THEN
        RSP:= 'ERROR: el camp ' || camp_nul || ' no pot ser nul'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
               
   WHEN E_NO_EXISTEIX THEN
        RSP:= 'ERROR: la lectura no existeix';      
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_NO_MOD THEN
        RSP:= 'ERROR: noms es permet modificar l''ltima lectura d''un comptador';
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);        
        
   WHEN E_LEC_VALOR THEN
        RSP:= 'ERROR: el valor del camp lectura_KWh ha de ser major o igual que ' || lec_anterior;   
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_LEC_DATA THEN
        RSP:= 'ERROR: el valor del camp data ha de ser posterior que ' || data_anterior;    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
        
   WHEN E_MES_NO_LEC THEN
        RSP:= 'ERROR: la nova data no s vlida. Ha d''existir almenys una lectura al mes anterior';    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
             
     
   WHEN OTHERS THEN
        ROLLBACK;      
        IF (SQLCODE = -02290) THEN 
          RSP:= 'ERROR: el valor del camp tipus ha de ser TELEMATIC/PRESENCIAL';
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
          
        ELSE
          err_msg := SUBSTR (SQLERRM, 1, 200);
          RSP:=    'ERROR: ' || err_msg;
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_LECTURA', param_entrada, 'RSP: '||RSP);
        END IF;
              
END SP_MODIFICACIO_LECTURA;