create or replace PROCEDURE SP_ALTA_PRODUCCIO_MENS (
   p_central_produccio IN  PRODUCCIO_MENSUAL.central_produccio%TYPE,
   p_data              IN  PRODUCCIO_MENSUAL.data%TYPE,
   p_energia_KWh       IN  PRODUCCIO_MENSUAL.energia_KWh%TYPE,
   RSP                 OUT VARCHAR2
)
IS
/******************************************************************************
   Nom:                 SP_ALTA_PRODUCCIO_MENS
   Propsit:            afegeix una entrada a la taula PRODUCCIO_MENSUAL
                        (lectura d'energa histrica produda)
                   
   Parmetres entrada:  p_central_produccio, p_data, p_energia_KWh             
   Parmetres sortida:  RSP
   Retorna:             'OK' si l'execuci ha finalitzat amb xit
                        'ERROR: tipus d'error' si ha fracassat
******************************************************************************/
param_entrada  VARCHAR2(500);
existeix     INTEGER;
err_msg      VARCHAR2 (200);
camp_nul     VARCHAR2(20);
lec_anterior NUMBER;
ultima_data  DATE;
produccio    NUMBER;
any_lectura  NUMBER;
E_NUL        EXCEPTION;
E_FK         EXCEPTION;
E_LEC_VALOR  EXCEPTION;
E_LEC_DATA   EXCEPTION;
E_MES_NO_LEC EXCEPTION;
BEGIN
   
   param_entrada:= 'p_central_produccio: '||p_central_produccio||
                 ', p_data: '||p_data||
                 ', p_energia_KWh: '||p_energia_KWh;
   
  --verifica que els camps obligatoris no siguin nuls
   IF (p_central_produccio IS NULL) THEN 
     camp_nul:= 'central_produccio';
     RAISE E_NUL;
   ELSIF (p_data IS NULL) THEN 
     camp_nul:= 'data';
     RAISE E_NUL;
   ELSIF (p_energia_KWh IS NULL) THEN 
     camp_nul:= 'energia_Kwh';
     RAISE E_NUL;   
   END IF; 

  --verifica que la central de producci existeix
   SELECT COUNT (*)
   INTO existeix
   FROM CENTRAL_PRODUCCIO
   WHERE codi = p_central_produccio ;

   IF (existeix = 0) THEN
     RAISE E_FK;
   END IF;      
   
  --calcula l'ultima data
   SELECT MAX(data) 
   INTO ultima_data 
   FROM PRODUCCIO_MENSUAL 
   WHERE central_produccio = p_central_produccio;
   
  --verifica que si no es tracta de la primera lectura, s'ha d'haver enregistrat
  --almenys una lectura al mes anterior.
   IF ( (EXTRACT(year FROM ultima_data) = EXTRACT(year FROM p_data) 
         AND EXTRACT(month FROM ultima_data) < EXTRACT(month FROM p_data)-1) 
         OR
        (EXTRACT(year FROM ultima_data) < EXTRACT(year FROM p_data)
         AND (EXTRACT(month FROM ultima_data)<>12 OR EXTRACT(month FROM p_data) <> 1 
              OR EXTRACT(year FROM p_data) > EXTRACT(year FROM ultima_data)+1))
      ) THEN
        RAISE E_MES_NO_LEC;
   END IF; 
    
  --calcula lectura anterior
   SELECT MAX(energia_KWh) 
   INTO lec_anterior 
   FROM PRODUCCIO_MENSUAL 
   WHERE central_produccio = p_central_produccio;
   
   --verifica que la lectura sigui la ms recent
   IF (p_data <= ultima_data)  THEN
     RAISE E_LEC_DATA;
   END IF;  
   
  --verifica que la lectura sigui major o igual que l'anterior
   IF (p_energia_kwh < lec_anterior)  THEN
     RAISE E_LEC_VALOR;
   END IF; 

  --calcula la produccio fins a la data de la lectura respecte a l'ultima lectura.
   SELECT p_energia_KWh - NVL( MAX(energia_KWh), 0)
   INTO produccio
   FROM PRODUCCIO_MENSUAL
   WHERE central_produccio = p_central_produccio;
   
  -- inserta la nova lectura
   INSERT INTO PRODUCCIO_MENSUAL (central_produccio, data, energia_KWh, produccio_kwh )
   VALUES (p_central_produccio, p_data, p_energia_KWh, produccio);
   
  -- finalitza amb xit
   RSP:= 'OK';
   SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
   
   
   
  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST5 ---
  ------------------------------------------------------------------------------

  --calcula l'any de la producci
   any_lectura:= EXTRACT(year FROM p_data);
   
  -- calcula si existeix una entrada a la taula per l'any de la produccio
   SELECT COUNT (*)
   INTO existeix
   FROM EST5_CPROD_MENYS_30
   WHERE anyo = any_lectura;
   
  --Si no existeix l'entrada, l'afegeix a EST5 
   IF (existeix = 0) THEN 
     SP_ALTA_EST5(any_lectura);
    
  --Si existeix actualitza la entrada.
   ELSE
   SP_MODIFICACIO_EST5(any_lectura);

   END IF;
   
   
   
   
   
   
   
   COMMIT;
EXCEPTION
   WHEN E_NUL THEN
        RSP:= 'ERROR: el camp ' || camp_nul || ' no pot ser nul'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
               
   WHEN E_FK THEN
        RSP:= 'ERROR: la central no existeix';    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);

   WHEN E_LEC_DATA THEN
        RSP:= 'ERROR: el valor del camp data ha de ser posterior que ' || ultima_data; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
  
   WHEN E_LEC_VALOR THEN
        RSP:= 'ERROR: el valor del camp energia_KWh ha de ser major o igual que ' || lec_anterior;   
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        
   WHEN E_MES_NO_LEC THEN
        RSP:= 'ERROR: no es pot enregistrar la producci. Ha d''existir almenys una lectura al mes anterior';      
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
     
   WHEN OTHERS THEN
        ROLLBACK;
        IF (SQLCODE = -00001) THEN 
          RSP:= 'ERROR: ja existeix una lectura amb mateix: central_produccio, data';        
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
          
        ELSE
          err_msg := SUBSTR (SQLERRM, 1, 200);
          RSP:=    'ERROR: ' || err_msg;
          SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_ALTA_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        END IF;
              
END SP_ALTA_PRODUCCIO_MENS;