create or replace PROCEDURE SP_MODIFICACIO_PRODUCCIO_MENS (
   p_codi              IN  PRODUCCIO_MENSUAL.codi%TYPE,
   p_data              IN  PRODUCCIO_MENSUAL.data%TYPE,
   p_energia_KWh       IN  PRODUCCIO_MENSUAL.energia_KWh%TYPE,
   RSP                 OUT VARCHAR2
)
IS
/******************************************************************************
   Nom:                 SP_MODIFICACIO_PRODUCCIO_MENS
   Propsit:            modifica les dades d'una lectura a la taula PRODUCCIO_MENSUAL
                        si es tracta de la lectura ms recent.
                        (lectura d'energa histrica produda)
                   
   Parmetres entrada:  p_codi, p_data, p_energia_KWh            
   Parmetres sortida:  RSP
   Retorna:             'OK' si l'execuci ha finalitzat amb xit
                        'ERROR: tipus d'error' si ha fracassat
******************************************************************************/
param_entrada  VARCHAR2(500);
existeix       INTEGER;
err_msg        VARCHAR2 (200);
camp_nul       VARCHAR2(20);
codi_central   NUMBER;
data_lectura   DATE;
any_lectura    NUMBER;
ultima_data    DATE;
data_anterior  DATE;
lec_anterior   NUMBER;
produccio      NUMBER;
E_NUL          EXCEPTION;
E_NO_EXISTEIX  EXCEPTION;
E_NO_ULTIMA    EXCEPTION;
E_LEC_VALOR    EXCEPTION;
E_LEC_DATA     EXCEPTION;
E_MES_NO_LEC   EXCEPTION;

BEGIN
   
   param_entrada:= 'p_codi: '||p_codi||
                 ', p_data: '||p_data||
                 ', p_energia_KWh: '||p_energia_KWh;
   
  --verifica que els camps obligatoris no siguin nuls
   IF (p_codi IS NULL) THEN 
     camp_nul:= 'codi';
     RAISE E_NUL;
   ELSIF (p_data IS NULL) THEN 
     camp_nul:= 'data';
     RAISE E_NUL;
   ELSIF (p_energia_KWh IS NULL) THEN 
     camp_nul:= 'energia_Kwh';
     RAISE E_NUL;   
   END IF; 

  --verifica que la lectura existeix
   SELECT COUNT (*)
   INTO existeix
   FROM PRODUCCIO_MENSUAL
   WHERE codi = p_codi ;

   IF (existeix = 0) THEN
     RAISE E_NO_EXISTEIX;
   END IF;      

  --calcula la central i la data de la lectura existent
   SELECT central_produccio, data 
   INTO codi_central, data_lectura
   FROM PRODUCCIO_MENSUAL 
   WHERE codi = p_codi;  

  --calcula l'ltima data
   SELECT MAX(data)
   INTO ultima_data
   FROM PRODUCCIO_MENSUAL
   WHERE central_produccio = codi_central;
    
  --verifica que es modifica l'ltima lectura de la central
   IF ( data_lectura <> ultima_data) THEN 
     RAISE E_NO_ULTIMA;
   END IF;
  
  --calcula lectura i data de la lectura anterior 
   SELECT MAX(energia_KWh), MAX(data)  
   INTO lec_anterior, data_anterior 
   FROM PRODUCCIO_MENSUAL 
   WHERE central_produccio = codi_central AND data < data_lectura;
   
  --verifica que la lectura sigui major o igual que l'anterior
   IF (p_energia_kwh < lec_anterior)  THEN
     RAISE E_LEC_VALOR;
   END IF;   
      
   --verifica que la nova data de lectura sigui la ms recent
   IF (p_data <= data_anterior)  THEN
     RAISE E_LEC_DATA;
   END IF;    
        

  --verifica que si no es tracta de la primera lectura, s'ha d'haver enregistrat
  --almenys una lectura al mes anterior.
   IF ( (EXTRACT(year FROM data_anterior) = EXTRACT(year FROM p_data) 
         AND EXTRACT(month FROM data_anterior) < EXTRACT(month FROM p_data)-1) 
         OR
        (EXTRACT(year FROM data_anterior) < EXTRACT(year FROM p_data)
         AND (EXTRACT(month FROM data_anterior)<>12 OR EXTRACT(month FROM p_data) <> 1 
              OR EXTRACT(year FROM p_data) > EXTRACT(year FROM data_anterior)+1))
      ) THEN
        RAISE E_MES_NO_LEC;
   END IF; 

  --calcula la produccio fins a la data respecte a l'ultima lectura.
   SELECT p_energia_KWh - NVL(energia_KWh, 0)
   INTO produccio
   FROM PRODUCCIO_MENSUAL
   WHERE central_produccio=codi_central AND data = data_anterior ;           
        
        
  -- modifica les dades de la lectura
   UPDATE PRODUCCIO_MENSUAL 
   SET data = p_data,
       energia_KWh = p_energia_kwh,
       produccio_kwh = produccio
   WHERE codi = p_codi;   
   
  -- finalitza amb xit
   RSP:= 'OK';
   SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);


  ------------------------------------------------------------------------------
                    -- ACTUALITZA EST5 ---
  ------------------------------------------------------------------------------
  
    --calcula l'any de la produccio
   any_lectura:= EXTRACT(year FROM p_data);
   
  --Actualitza l'entrada de l'any a EST5.
   SP_MODIFICACIO_EST5(any_lectura);





   COMMIT;
EXCEPTION
   WHEN E_NUL THEN
        RSP:= 'ERROR: el camp ' || camp_nul || ' no pot ser nul'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
               
   WHEN E_NO_EXISTEIX THEN
        RSP:= 'ERROR: la lectura no existeix';      
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        
   WHEN E_NO_ULTIMA THEN
        RSP:= 'ERROR: noms es permet modificar l''ltima lectura de la central'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        
   WHEN E_LEC_VALOR THEN
        RSP:= 'ERROR: el valor del camp energia_KWh ha de ser major o igual que ' || lec_anterior;   
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        
   WHEN E_LEC_DATA THEN
        RSP:= 'ERROR: el valor del camp data ha de ser posterior que ' || data_anterior;    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
        
   WHEN E_MES_NO_LEC THEN
        RSP:= 'ERROR: la nova data no s vlida. Ha d''existir almenys una lectura al mes anterior';    
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);
     
   WHEN OTHERS THEN
        ROLLBACK;      
        err_msg := SUBSTR (SQLERRM, 1, 200);
        RSP:=    'ERROR: ' || err_msg;
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'SP_MODIFICACIO_PRODUCCIO_MENS', param_entrada, 'RSP: '||RSP);

END SP_MODIFICACIO_PRODUCCIO_MENS;