

create or replace TYPE TY_R6a AS OBJECT (
  codi_contracte           NUMBER, 
  potencia_KW              NUMBER,
  percent_respecte_mitjana NUMBER
);
/
   
 
create or replace TYPE TA_R6a AS TABLE OF TY_R6a;             
/


create or replace FUNCTION FU_COMPT_MES_80_CONSUM (
   p_localitat IN  COMPTADOR.localitat%TYPE,
   p_data      IN  LECTURA.data%TYPE
)
RETURN TA_R6a
IS
/******************************************************************************
   Nom:                 FU_COMPT_MES_80_CONSUM
   Propsit:            Donada una ciutat i una data com a parmetres, obt el 
                        llistat de tots els comptadors on el consum mensual de 
                        la data indicada ha superat el 80% del consum mitj de 
                        tots els comptadors de la ciutat en aquell mateix perode 
                        de temps. 
                   
   Parmetres entrada:  -	p_localitat  VARCHAR2(20):  localitat o poblaci on es
                          vol consultar la informaci.
                        -	p_data  DATE: data en que es vol fer la consulta. 

   Parmetres sortida:  -	RSP 
   Retorna:             -	si l'execuci ha finalitzat amb xit retorna el resultat 
                          de la consulta especificant :
                          	el codi de contracte.     
                          	potncia mxima contractada. 
                          	percentatge de consum respecte al consum mitja.

                          NOTA: Si no es troba cap resultat la taula retornada 
                          estar buida.

                        - 'ERROR: tipus d'error' si ha fracassat       
******************************************************************************/
param_entrada  VARCHAR2(500);
err_msg        VARCHAR2 (200);
camp_nul       VARCHAR2(20);
mitjana        NUMBER;
contr          NUMBER;
potencia       NUMBER;
consum         NUMBER;
i              INTEGER:=1;
RSP	           VARCHAR2(500);
resultat       TA_R6a:=TA_R6a();
E_NUL          EXCEPTION;
E_MES          EXCEPTION;

type r_cursor IS REF CURSOR;
CUR r_cursor;

BEGIN
   
   param_entrada:= 'p_localitat: '||p_localitat||
                 ', p_data: '||p_data;
                 
  --verifica que els camps no siguin nuls
   IF (p_localitat IS NULL) THEN 
     camp_nul:= 'localitat';
     RAISE E_NUL;
   ELSIF (p_data IS NULL) THEN 
     camp_nul:= 'data';
     RAISE E_NUL;     
   END IF;  
   
  --verifica que la consulta no sigui amb data del mes actual, ja que pot ser que no
  --es tinguin totes les lectures de tots els comptadors.
   IF (EXTRACT (month FROM p_data) = EXTRACT (month FROM SYSDATE) ) THEN 
   RAISE E_MES;
   END IF;
   
  --Calcula la mitjana 
   SELECT ROUND( AVG(SUM(l.consum_KWh)), 2)
   INTO mitjana   
   FROM lectura l JOIN comptador c ON  l.comptador = c.codi
   WHERE localitat = p_localitat AND data BETWEEN TRUNC(p_data, 'month') AND  p_data
   GROUP BY comptador; 
    
   OPEN CUR FOR SELECT ct.codi, ct.potencia_kw, SUM(l.consum_kwh) 
                FROM ( contracte ct JOIN lectura l ON ct.comptador = l.comptador ) JOIN comptador c ON ct.comptador = c.codi
                WHERE localitat = p_localitat AND data BETWEEN TRUNC(p_data, 'month') AND  p_data
                GROUP BY ct.codi, ct.potencia_kw
                ORDER BY 3;    
                               
   LOOP
     FETCH CUR INTO contr, potencia, consum;
     EXIT WHEN CUR%NOTFOUND;
     IF (consum >= mitjana * 0.8) THEN 
       resultat.extend;
       resultat(i):= TY_R6a(contr, potencia, round((consum/mitjana)*100, 2));
       i:= i+1;
     END IF;
   END LOOP;  
   CLOSE CUR;  
   
  -- finalitza amb xit
   RSP:= 'OK';
   SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'FU_COMPT_MES_80_CONSUM', param_entrada, 'RSP: '||RSP);


  --retorna resultat;
   RETURN resultat;
EXCEPTION
   WHEN E_NUL THEN
        RSP:= 'ERROR: el camp ' || camp_nul || ' no pot ser nul'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'FU_COMPT_MES_80_CONSUM', param_entrada, 'RSP: '||RSP);
        DBMS_OUTPUT.PUT_LINE(RSP);
        return null;

   WHEN E_MES THEN
        RSP:= 'ERROR: no es permet fer la consulta sobre el mes actual'; 
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'FU_COMPT_MES_80_CONSUM', param_entrada, 'RSP: '||RSP);
        DBMS_OUTPUT.PUT_LINE(RSP);
        return null;

   WHEN OTHERS THEN
        err_msg := SUBSTR (SQLERRM, 1, 200);
        RSP:=    'ERROR: ' || err_msg;
        SP_ALTA_LOGS(to_char(sysdate,'DD/MM/YYYY HH24:MI:SS'), 'FU_COMPT_MES_80_CONSUM', param_entrada, 'RSP: '||RSP);
        DBMS_OUTPUT.PUT_LINE(RSP);
        return null;
END;